#!/usr/bin/env python3
#	pysuchsel - Create Suchsel word puzzles from Python
#	Copyright (C) 2019-2023 Johannes Bauer
#
#	This file is part of pysuchsel.
#
#	pysuchsel is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pysuchsel is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import sys
import pysuchsel
from .MultiCommand import MultiCommand
from .ActionSuchselCrossword import ActionSuchselCrossword
from .ActionSolutionWord import ActionSolutionWord
from .ActionCrypto import ActionCrypto
from .ActionFontTest import ActionFontTest

def main():
	mc = MultiCommand(trailing_text = "version: pysuchsel v%s" % (pysuchsel.VERSION))

	def genparser(parser):
		parser.add_argument("-f", "--fill-rule", choices = [ "en", "de" ], default = "en", help = "Distribute empty spaces with this alphabet and natural language frequency distribution. I.e., every letter occurs with the same probability or is the frequency that is also commonly found in the natural language used. Can be one of %(choices)s, defaults to %(default)s.")
		parser.add_argument("--uniform-distribution", action = "store_true", help = "Instead of choosing natural language distribution, distribute space letters uniformly. Each letter will appear with the same probability.")
		parser.add_argument("-x", "--width", metavar = "width", type = int, default = 15, help = "Choose this width for the suchsel. Defaults to %(default)d spaces.")
		parser.add_argument("-y", "--height", metavar = "height", type = int, default = 20, help = "Choose this height for the suchsel. Defaults to %(default)d spaces.")
		parser.add_argument("-p", "--placement", choices = [ "lr", "tb", "rl", "bt", "dbr", "dtr", "dbl", "dtl" ], action = "append", default = [ ], help = "Defines the placement rule of words within the suchsel. Can be specified multiple times and accepts %(choices)s as option. By default tb and lr is used (top -> bottom and left -> right). Choices beginning with 'd' mean diagonal (diagonal to bottom right/bottom left/top right/top left).")
		parser.add_argument("-c", "--contiguous", action = "store_true", help = "Try to create a contiguous Suchsel, i.e., where some letters overlap.")
		parser.add_argument("--place-attempts", metavar = "cnt", type = int, default = 500, help = "Placing words is non-deterministic. This increases the amounts of attempts for placing a word before giving up. Longer might yield better results, but also takes longer.")
		parser.add_argument("-a", "--creation-attempts", metavar = "cnt", type = int, default = 1, help = "Sometimes, not all words can be placed. This gives the number of attempts that creation of the Suchsel/cross word puzzle is re-attempted before giving up.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increase verbosity. Can be specified multiple times.")
		parser.add_argument("infile", metavar = "infile", help = "Input filename that contains all words separated by newlines.")
		parser.add_argument("outfile", metavar = "outfile", help = "Output SVG file to create.")
	mc.register("suchsel", "Create a Suchsel word puzzle", genparser, action = ActionSuchselCrossword)
	mc.register("crossword", "Create a crossword puzzle", genparser, action = ActionSuchselCrossword)

	def genparser(parser):
		parser.add_argument("--place-attempts", metavar = "cnt", type = int, default = 500, help = "Placing words is non-deterministic. This increases the amounts of attempts for placing a word before giving up. Longer might yield better results, but also takes longer.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increase verbosity. Can be specified multiple times.")
		parser.add_argument("infile", metavar = "infile", help = "Input filename that contains all words separated by newlines.")
		parser.add_argument("outfile", metavar = "outfile", help = "Output SVG file to create.")
		parser.add_argument("solword", metavar = "word", help = "Solution word to search.")
	mc.register("solword", "Create a solution word puzzle", genparser, action = ActionSolutionWord)

	def genparser(parser):
		parser.add_argument("-a", "--alphabet", choices = [ "alpha", "math", "graph", "zodiac", "chess", "runes" ], action = "append", default = [ ], required = True, help = "Name of the ciphertext alphabet(s) to use. Can be specified multiple times, can be any of %(choices)s. Must be given at least once.")
		parser.add_argument("-r", "--reveal", metavar = "letters", default = "ERNSTL", help = "Letters to initially reveal. Defaults to %(default)s.")
		parser.add_argument("-w", "--solution-word", metavar = "word", help = "When puzzle should contain a final solution word, this parameter sets it.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increase verbosity. Can be specified multiple times.")
		parser.add_argument("infile", metavar = "infile", help = "Input filename that contains all lines separated by newlines.")
		parser.add_argument("outfile", metavar = "outfile", help = "Output SVG file to create.")
	mc.register("crypto", "Create crypto word puzzle", genparser, action = ActionCrypto)

	def genparser(parser):
		parser.add_argument("-s", "--sort", action = "store_true", help = "Sort the alphabets before printing")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increase verbosity. Can be specified multiple times.")
		parser.add_argument("outfile", metavar = "outfile", help = "Output SVG file to create.")
	mc.register("fonttest", "Test the fonts for crypto word puzzles", genparser, action = ActionFontTest)

	return mc.run(sys.argv[1:])

if __name__  == "__main__":
	main()



#	pysuchsel - Create Suchsel word puzzles from Python
#	Copyright (C) 2019-2023 Johannes Bauer
#
#	This file is part of pysuchsel.
#
#	pysuchsel is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pysuchsel is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import json
import pkgutil
from pysvgedit import SVGDocument, SVGRect, Vector2D, Convenience as svgc
from .BaseAction import BaseAction

class ActionFontTest(BaseAction):
	def run(self):
		svg = SVGDocument.new()
		crypto = json.loads(pkgutil.get_data("pysuchsel", "definitions.json"))["crypto"]
		size = 20
		for (y, (name, alphabets)) in enumerate(crypto.items()):
			svgc.text(svg, text = name, x = -100, y = size * y, width = 90, height = size, halign = "right")
			alphabet = "".join(alphabets)
			if self._args.sort:
				alphabet = sorted(set(alphabet))

			yshift = 4
			for (x, letter) in enumerate(alphabet):
				svg.add(SVGRect.new(pos = Vector2D(size * x, size * y), extents = Vector2D(size, size)))
				svgc.text(svg, text = letter, x = size * x, y = size * y + yshift, width = size, height = size - yshift, halign = "center")
		svgc.autosize(svg)
		svg.writefile(self._args.outfile)



#	pysuchsel - Create Suchsel word puzzles from Python
#	Copyright (C) 2019-2023 Johannes Bauer
#
#	This file is part of pysuchsel.
#
#	pysuchsel is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pysuchsel is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import random
from pysvgedit import SVGDocument, SVGGroup, SVGRect, Vector2D, Convenience as svgc

class VoidPlaceholder():
	def __eq__(self, other):
		return other.__class__ == self.__class__

	def __neq__(self, other):
		return not (self == other)

	def __str__(self):
		return "."

class ArrowMarker():
	def __init__(self, marking, direction):
		self._marking = marking
		self._direction = direction

	@property
	def marking(self):
		return self._marking

	def __str__(self):
		return {
			"left":		"<",
			"right":	">",
			"down":		"v",
		}.get(self._direction, "?")

class Suchsel():
	def __init__(self, width, height, placement, attempts, is_crossword = False):
		self._width = width
		self._height = height
		self._placement = placement
		self._attempts = attempts
		self._is_crossword = is_crossword
		self._grid = { }
		self._fillers_at = set()

	def _rulerange(self, origin_x, origin_y, rulename):
		(x, y) = (origin_x, origin_y)
		while True:
			yield (x, y)
			if rulename == "lr":
				x += 1
			elif rulename == "tb":
				y += 1
			elif rulename == "dbr":
				x += 1
				y += 1
			elif rulename == "dbl":
				x -= 1
				y += 1
			elif rulename == "dtr":
				x += 1
				y -= 1
			elif rulename == "dtl":
				x -= 1
				y -= 1
			else:
				raise NotImplementedError(rulename)

	def _adjacent_fields(self, x, y, rulename):
		if rulename in [ "tb", "bt" ]:
			yield (x - 1, y)
			yield (x + 1, y)
		elif rulename in [ "lr", "rl" ]:
			yield (x, y - 1)
			yield (x, y + 1)
		else:
			raise NotImplementedError(rulename)

	def _find_rule(self, word, crossword_marker = None):
		rule = self._placement.event()
		if crossword_marker is not None:
			direction = {
				"lr":	"right",
				"rl":	"left",
				"tb":	"down",
				"bt":	"up",
				"dbr":	"down-right",
				"dbl":	"down-left",
				"dtr":	"up-right",
				"dtl":	"up-left",
			}[rule]
			arrow_marker = ArrowMarker(marking = crossword_marker, direction = direction)
			word = [ arrow_marker ] + list(word) + [ VoidPlaceholder() ]

		if rule in [ "lr", "rl" ]:
			req_width = len(word)
			req_height = 1
		elif rule in [ "tb", "bt" ]:
			req_width = 1
			req_height = len(word)
		elif rule in [ "dbr", "dtl", "dtr", "dbl" ]:
			req_width = len(word)
			req_height = len(word)
		else:
			raise NotImplementedError(rule)

		if rule in [ "rl", "bt" ]:
			word = list(reversed(word))
			if rule == "rl":
				rule = "lr"
			elif rule == "bt":
				rule = "tb"
		return (word, rule, req_width, req_height)

	def _attempt_place(self, word, must_be_contiguous = False, crossword_marker = None):
		(word, rule, req_width, req_height) = self._find_rule(word, crossword_marker = crossword_marker)

		max_x = self._width - req_width
		max_y = self._height - req_height
		if (max_x < 0) or (max_y < 0):
			# Word does not fit with this rule, abort.
			return False

		src_x = random.randint(0, max_x)
		src_y = random.randint(0, max_y)

		if rule in [ "dtl", "dbl" ]:
			src_x += len(word) - 1
		if rule in [ "dtr", "dtl" ]:
			src_y += len(word) - 1

		contiguous_letters = 0
		for (want_place, (x, y)) in zip(word, self._rulerange(src_x, src_y, rule)):
			present = self._grid.get((x, y))
			if (present == want_place) and isinstance(present, str):
				# We count overlapping letters, but not overlapping
				# VoidPlacerholders
				contiguous_letters += 1
			if (present is not None) and (present != want_place):
				# Letter already occupied with different letter than we would
				# like there
				return False

			if (present is None) and (crossword_marker is not None):
				# Field is empty, check adjacent fields for emptyness if this
				# is a crossword
				for (adjx, adjy) in self._adjacent_fields(x, y, rule):
					adjacent_content = self._grid.get((adjx, adjy))
					if (adjacent_content is not None) and not isinstance(adjacent_content, VoidPlaceholder):
						# There's a letter or arrowfield in there, that's
						# forbidden
						return False

		if must_be_contiguous and (contiguous_letters == 0):
			return False

		# All letters fit!
		for (want_place, (x, y)) in zip(word, self._rulerange(src_x, src_y, rule)):
			self._grid[(x, y)] = want_place
		return True

	def place(self, word, contiguous = False):
		# First try contiguous placement
		if contiguous and (len(self._grid) > 0):
			for i in range(self._attempts):
				if self._attempt_place(word, must_be_contiguous = True):
					return True
		for i in range(self._attempts):
			if self._attempt_place(word):
				return True
		return False

	def place_crossword(self, word, crossword_marker):
		contiguous = (len(self._grid) > 0)
		for i in range(self._attempts):
			if self._attempt_place(word, must_be_contiguous = contiguous, crossword_marker = crossword_marker):
				return True
		return False

	def fill(self, filler):
		for y in range(self._height):
			for x in range(self._width):
				pos = (x, y)
				if pos not in self._grid:
					self._grid[pos] = filler.get()
					self._fillers_at.add(pos)

	def dump(self):
		print("+-" + "-" * (2 * self._width) + "-+")
		for y in range(self._height):
			line = [ ]
			for x in range(self._width):
				letter = self._grid.get((x, y), " ")
				line.append(str(letter))
			print("| " + (" ".join(line)) + "  |")
		print("+-" + ("-" * (2 * self._width)) + "-+")

	def write_svg(self, output_filename):
		svg = SVGDocument.new()

		grid_layer = svg.add(SVGGroup.new(is_layer = True))
		grid_layer.label = "Grid"
		if self._is_crossword:
			solution_layer = svg.add(SVGGroup.new(is_layer = True))
			solution_layer.label = "Solution"
		else:
			filler_layer = svg.add(SVGGroup.new(is_layer = True))
			filler_layer.label = "Filler letters"
			solution_layer = svg.add(SVGGroup.new(is_layer = True))
			solution_layer.label = "Solution normal"
			solution_highlight_layer = svg.add(SVGGroup.new(is_layer = True))
			solution_highlight_layer.label = "Solution highlighted"

		size = 20
		yoffset = 4
		for y in range(self._height):
			for x in range(self._width):
				pos = (x, y)
				letter = self._grid.get(pos)
				is_filler = pos in self._fillers_at
				if isinstance(letter, str):
					grid_layer.add(SVGRect.new(pos = size * Vector2D(x, y), extents = Vector2D(size, size)))

					if is_filler:
						# Random filler letter
						svgc.text(filler_layer, pos = Vector2D(size * x, size * y + yoffset), extents = Vector2D(size, size - yoffset), text = letter, halign = "center")

					if not is_filler:
						# Solution letter
						svgc.text(solution_layer, pos = Vector2D(size * x, size * y + yoffset), extents = Vector2D(size, size - yoffset), text = letter, halign = "center")

						if not self._is_crossword:
							# Additionally, put in bold and highlight in the solution highlight layer
							rect = solution_highlight_layer.add(SVGRect.new(pos = size * Vector2D(x, y), extents = Vector2D(size, size)))
							rect.style["fill"] = "#f1c40f"
							svgc.text(solution_highlight_layer, pos = Vector2D(size * x, size * y + yoffset), extents = Vector2D(size, size - yoffset), text = letter, halign = "center", attribute = "bold")
				elif isinstance(letter, ArrowMarker):
					rect = grid_layer.add(SVGRect.new(pos = size * Vector2D(x, y), extents = Vector2D(size, size)))
					rect.style["stroke"] = "#3498db"
					svgc.text(grid_layer, pos = Vector2D(size * x, size * y + yoffset), extents = Vector2D(size, size - yoffset), text = str(letter.marking), halign = "center")

		svgc.autosize(svg)
		svg.writefile(output_filename)



#	pysuchsel - Create Suchsel word puzzles from Python
#	Copyright (C) 2019-2021 Johannes Bauer
#
#	This file is part of pysuchsel.
#
#	pysuchsel is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pysuchsel is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#	Johannes Bauer <JohannesBauer@gmx.de>

class PuzzleNotSolvableException(Exception): pass



#	pysuchsel - Create Suchsel word puzzles from Python
#	Copyright (C) 2019-2023 Johannes Bauer
#
#	This file is part of pysuchsel.
#
#	pysuchsel is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pysuchsel is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from .BaseAction import BaseAction
from .SolutionWordPuzzle import SolutionWordPuzzle, PuzzleNotSolvableException
from .Tools import Tools

class ActionSolutionWord(BaseAction):
	def run(self):
		word_list = Tools.read_file(self._args.infile)
		swp = SolutionWordPuzzle(word_list = word_list, solution_word = self._args.solword)
		solution = None
		for i in range(self._args.place_attempts):
			try:
				solution = swp.find_solution()
			except PuzzleNotSolvableException:
				continue
		if solution is None:
			print("Could not find a solution for this word puzzle, giving up.")
			return 1

		if self._args.verbose >= 1:
			indent = max(index for (word, index) in solution)
			print("   " + (" " * indent) + "#")
			for (word_no, (word, letter_index)) in enumerate(solution, 1):
				word_indent = " " * (indent - letter_index)
				print("%2d %s%s" % (word_no, word_indent, word))
		swp.write_svg(self._args.outfile)
		return 0



#	pysuchsel - Create Suchsel word puzzles from Python
#	Copyright (C) 2019-2023 Johannes Bauer
#
#	This file is part of pysuchsel.
#
#	pysuchsel is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pysuchsel is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from .BaseAction import BaseAction
from .CryptoPuzzle import CryptoPuzzle
from .Tools import Tools

class ActionCrypto(BaseAction):
	def run(self):
		if len(self._args.alphabet) == 0:
			raise Exception("No alphabet given on command line.")
		plain_lines = Tools.read_file(self._args.infile)
		cp = CryptoPuzzle(plain_lines = plain_lines, alphabet_names = self._args.alphabet, reveal_letters = self._args.reveal, crypto_solution = self._args.solution_word)
		if self._args.verbose >= 1:
			cp.dump()
		cp.write_svg(self._args.outfile)



#	pysuchsel - Create Suchsel word puzzles from Python
#	Copyright (C) 2019-2023 Johannes Bauer
#
#	This file is part of pysuchsel.
#
#	pysuchsel is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pysuchsel is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import json
import pkgutil
from .RandomDist import RandomDist

class Alphabet():
	def __init__(self, language, uniform_distribution = False):
		distributions = json.loads(pkgutil.get_data("pysuchsel", "definitions.json"))["distributions"]
		distribution = distributions[language]

		if uniform_distribution:
			self._dist = RandomDist({ letter: 1 for letter in distribution.keys() })
		else:
			self._dist = RandomDist({ letter: round(10000 * probability) for (letter, probability) in distribution.items() })

	def get(self):
		return self._dist.event()



#	pysuchsel - Create Suchsel word puzzles from Python
#	Copyright (C) 2019-2023 Johannes Bauer
#
#	This file is part of pysuchsel.
#
#	pysuchsel is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pysuchsel is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from .BaseAction import BaseAction
from .RandomDist import RandomDist
from .Suchsel import Suchsel
from .Tools import Tools
from .Alphabet import Alphabet

class ActionSuchselCrossword(BaseAction):
	def _get_placement_rule(self):
		if len(self._args.placement) == 0:
			plcrule = RandomDist({
				"lr":	1,
				"tb":	1,
			})
		else:
			plcrule = RandomDist({ name: 1 for name in self._args.placement })
		return plcrule

	def _attempt_placement(self, plcrule):
		self._unplaced_words = [ ]
		self._placed_words = { }
		next_id = 1
		self._suchsel = Suchsel(self._args.width, self._args.height, plcrule, attempts = self._args.place_attempts, is_crossword = (self._cmd == "crossword"))
		for word in self._words:
			if self._cmd == "suchsel":
				placed = self._suchsel.place(word, contiguous = self._args.contiguous)
			elif self._cmd == "crossword":
				placed = self._suchsel.place_crossword(word, crossword_marker = next_id)
				if placed:
					self._placed_words[next_id] = word
					next_id += 1
			else:
				raise NotImplementedError(self._args.mode)
			if not placed:
				self._unplaced_words.append(word)
		return len(self._unplaced_words) == 0

	def run(self):
		self._words = Tools.read_file(self._args.infile, shuffle = True)
		plcrule = self._get_placement_rule()

		for creation_attempt in range(self._args.creation_attempts):
			if self._attempt_placement(plcrule):
				break

		for unplaced_word in self._unplaced_words:
			print("Warning: could not place word \"%s\"." % (unplaced_word))

		if self._args.verbose >= 1:
			self._suchsel.dump()

		if self._cmd == "suchsel":
			filler = Alphabet(self._args.fill_rule, uniform_distribution = self._args.uniform_distribution)
			self._suchsel.fill(filler)
			if self._args.verbose >= 2:
				self._suchsel.dump()

		self._suchsel.write_svg(self._args.outfile)

		if self._cmd == "crossword":
			for (word_id, word) in sorted(self._placed_words.items()):
				print("%2d: %s" % (word_id, word))



#	pysuchsel - Create Suchsel word puzzles from Python
#	Copyright (C) 2019-2023 Johannes Bauer
#
#	This file is part of pysuchsel.
#
#	pysuchsel is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pysuchsel is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import random
import json
import pkgutil
from pysvgedit import SVGDocument, SVGGroup, SVGRect, Vector2D, Convenience as svgc
from .Exceptions import PuzzleNotSolvableException

class CryptoPuzzle():
	def __init__(self, plain_lines, alphabet_names, reveal_letters, crypto_solution = None):
		self._plain_lines = plain_lines
		self._crypto_solution = crypto_solution
		self._reveal_letters = set(reveal_letters)
		crypto_alphabet_names = json.loads(pkgutil.get_data("pysuchsel", "definitions.json"))["crypto"]

		crypto_chars = set()
		for alphabet in alphabet_names:
			crypto_chars |= set("".join(crypto_alphabet_names[alphabet]))
		crypto_chars = list(crypto_chars)
		random.shuffle(crypto_chars)

		plain_chars = set("".join(word for word in self._plain_lines))
		if " " in plain_chars:
			plain_chars.remove(" ")
		if len(plain_chars) > len(crypto_chars):
			raise PuzzleNotSolvableException("Alphabet too small, cannot generate puzzle.")
		self._key = { plain_char: crypto_char for (plain_char, crypto_char) in zip(plain_chars, crypto_chars) }

		if self._crypto_solution is not None:
			uncovered_chars = set(crypto_solution) - plain_chars
			if len(uncovered_chars) > 0:
				raise PuzzleNotSolvableException("Cannot produce solution word: %s missing" % (", ".join(sorted(uncovered_chars))))

	def dump(self):
		for line in self._plain_lines:
			show = [ ]
			for letter in line:
				if (letter == " ") or (letter in self._reveal_letters):
					show.append(letter)
				else:
					show.append("_")
			print("".join(show))


	def write_svg(self, output_filename):
		svg = SVGDocument.new()
		cipher_layer = svg.add(SVGGroup.new(is_layer = True))
		cipher_layer.label = "Ciphertext"

		initial_layer = svg.add(SVGGroup.new(is_layer = True))
		initial_layer.label = "Initially given"

		inference_layer = svg.add(SVGGroup.new(is_layer = True))
		inference_layer.label = "Initially inferrable"
		inference_layer.style.hide()

		solution_layer = svg.add(SVGGroup.new(is_layer = True))
		solution_layer.label = "Solution"
		solution_layer.style.hide()

		if self._crypto_solution is not None:
			final_ciphertext = svg.add(SVGGroup.new(is_layer = True))
			final_ciphertext.label = "Final ciphertext"

			final_plaintext = svg.add(SVGGroup.new(is_layer = True))
			final_plaintext.label = "Final solution"
			final_plaintext.style.hide()


		size = 20
		yoffset = 4
		line_height = 2 * size + 10

		revealed_letters = set()
		for (y, line) in enumerate(self._plain_lines):
			for (x, plain_letter) in enumerate(line):
				if plain_letter == " ":
					continue
				cipher_letter = self._key[plain_letter]
				cipher_layer.add(SVGRect.new(pos = Vector2D(size * x, line_height * y), extents = Vector2D(size, size)))
				cipher_layer.add(SVGRect.new(pos = Vector2D(size * x, line_height * y + size), extents = Vector2D(size, size)))
				svgc.text(cipher_layer, pos = Vector2D(size * x, line_height * y + yoffset), extents = Vector2D(size, size - yoffset), text = cipher_letter, halign = "center")
				svgc.text(solution_layer, pos = Vector2D(size * x, line_height * y + size + yoffset), extents = Vector2D(size, size - yoffset), text = plain_letter, halign = "center", attribute = "bold")
				if plain_letter in self._reveal_letters:
					svgc.text(inference_layer, pos = Vector2D(size * x, line_height * y + size + yoffset), extents = Vector2D(size, size - yoffset), text = plain_letter, halign = "center")
					if plain_letter not in revealed_letters:
						svgc.text(initial_layer, pos = Vector2D(size * x, line_height * y + size + yoffset), extents = Vector2D(size, size - yoffset), text = plain_letter, halign = "center")
						revealed_letters.add(plain_letter)

		y += 1
		half_height = size // 2
		if self._crypto_solution is not None:
			for (x, plain_letter) in enumerate(self._crypto_solution):
				cipher_letter = self._key[plain_letter]
				rect = final_ciphertext.add(SVGRect.new(pos = Vector2D(size * x, line_height * y + half_height), extents = Vector2D(size, size)))
				rect.style["fill"] = "#f1c40f"

				final_ciphertext.add(SVGRect.new(pos = Vector2D(size * x, line_height * y + size + half_height), extents = Vector2D(size, size)))
				svgc.text(final_ciphertext, pos = Vector2D(size * x, line_height * y + half_height + yoffset), extents = Vector2D(size, size - yoffset), text = cipher_letter, halign = "center")

				svgc.text(final_plaintext, pos = Vector2D(size * x, line_height * y + size + half_height + yoffset), extents = Vector2D(size, size - yoffset), text = plain_letter, halign = "center", attribute = "bold")

		svgc.autosize(svg)
		svg.writefile(output_filename)

if __name__ == "__main__":
	cp = CryptoPuzzle([ "THIS IS A", "SUPER SECRET", "MESSAGE" ], [ "alpha" ], "ERNSTL", crypto_solution = "TATA")
	cp.dump()
	cp.write_svg("x.svg")



#	pysuchsel - Create Suchsel word puzzles from Python
#	Copyright (C) 2019-2023 Johannes Bauer
#
#	This file is part of pysuchsel.
#
#	pysuchsel is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pysuchsel is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import random
from pysvgedit import SVGDocument, SVGGroup, SVGRect, Vector2D, Convenience as svgc
from .Exceptions import PuzzleNotSolvableException

class SolutionWordPuzzle():
	def __init__(self, word_list, solution_word):
		self._word_list = [ word.upper() for word in word_list ]
		self._solution_word = solution_word.upper()
		self._plausibilize()
		self._solution = None

	def _plausibilize(self):
		available_letters = set("".join(word for word in self._word_list))
		necessary_letters = set(self._solution_word)
		unavailable_letters = necessary_letters - available_letters
		if len(unavailable_letters) > 0:
			raise PuzzleNotSolvableException("Puzzle not solvable: letter(s) %s not contained." % (", ".join(sorted(unavailable_letters))))

	def find_solution(self):
		remaining_words = list(self._word_list)
		random.shuffle(remaining_words)
		solution_words = [ ]
		for letter in self._solution_word:
			for (index, candidate) in enumerate(remaining_words):
				if letter in candidate:
					break
			else:
				raise PuzzleNotSolvableException("Could not find word for letter '%s'." % (letter))
			remaining_words.pop(index)
			solution_words.append(candidate)

		solution = [ ]
		for (solution_word, solution_letter) in zip(solution_words, self._solution_word):
			occurrence_indices = set()
			for (index, letter) in enumerate(solution_word):
				if letter == solution_letter:
					occurrence_indices.add(index)
			chosen_index = random.choice(list(occurrence_indices))
			solution.append((solution_word, chosen_index))
		self._solution = solution
		return solution

	def write_svg(self, output_filename):
		svg = SVGDocument.new()
		grid_layer = svg.add(SVGGroup.new(is_layer = True))
		grid_layer.label = "Grid"
		solution_layer = svg.add(SVGGroup.new(is_layer = True))
		solution_layer.label = "Solution"

		size = 20
		yoffset = 4
		for (y, (word, letter_index)) in enumerate(self._solution):

			x_begin = -letter_index - 1
			rect = grid_layer.add(SVGRect.new(pos = size * Vector2D(x_begin, y), extents = Vector2D(size, size)))
			rect.style["fill"] = "#3498db"
			svgc.text(grid_layer, pos = Vector2D(size * x_begin, size * y + yoffset), extents = Vector2D(size, size - yoffset), text = str(y + 1), halign = "center")

			for (x_raw, letter) in enumerate(word):
				x = x_raw - letter_index

				if x == 0:
					rect = grid_layer.add(SVGRect.new(pos = size * Vector2D(x, y), extents = Vector2D(size, size)))
					rect.style["fill"] = "#f1c40f"
				else:
					grid_layer.add(SVGRect.new(pos = size * Vector2D(x, y), extents = Vector2D(size, size)))

				svgc.text(solution_layer, pos = Vector2D(size * x, size * y + yoffset), extents = Vector2D(size, size - yoffset), text = letter, halign = "center", attribute = "bold" if (x == 0) else "none")

		svgc.autosize(svg)
		svg.writefile(output_filename)


if __name__ == "__main__":
	swp = SolutionWordPuzzle([ "PADDELFISCH", "ITZIBITZI", "NUDELSALAT", "SOMMER", "SONNE", "HEITERKEIT" ], "MILK")
	print(swp.find_solution())



#!/usr/bin/python3
#
#	RandomDist - Generate random values of given distribution
#	Copyright (C) 2011-2023 Johannes Bauer
#
#	This file is part of pycommon.
#
#	pycommon is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pycommon is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pycommon; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>
#
#	File UUID 7e9a90a5-a1a1-4e49-a67c-5d935126abfe

import random

class RandomDist():
	def __init__(self, distribution):
		self._sum = 0
		self._values = [ ]
		for (key, value) in distribution.items():
			if value > 0:
				self._sum += value
				self._values.append((key, self._sum))

	def coinflip(self):
		return random.randint(0, 1) == 0

	def event(self):
		randval = random.random() * self._sum
		for (key, value) in self._values:
			if randval < value:
				return key
		return None

if __name__ == "__main__":
	rdist = RandomDist({
		"10_1":	10,
		"10_2":	10,
		"20":	20,
		"60":	60,
	})

	total = 300000
	events = { }
	for i in range(total):
		x = rdist.event()
		events[x] = events.get(x, 0) + 1
	for (event, cnt) in events.items():
		print("%-8s %7.4f%%" % (event, cnt / total * 100))


